==> entity_query_abm-1.0.68264.tar.gz: entity_query_abm-1.0.68264/entity_query/relation.py <==
#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from dataclasses import dataclass

import sqlalchemy.orm

from entity_read.entity import Entity
from .order import Order
from .expression import Expression
from .query_token import QueryToken


@dataclass(frozen=True, kw_only=True, repr=False)
class Relation(QueryToken):
    key: str

    def get_related_entity(self, entity_type: type[Entity]) -> type[Entity]:
        relation = self._get_relation(entity_type)
        result = relation.entity.entity
        return result

    def get_local_column_name_to_remote(self, entity_type: type[Entity]) -> dict[str, str]:
        relation = self._get_relation(entity_type)
        return {local_column.name: remote_column.name for local_column, remote_column in relation.local_remote_pairs}

    def get_remote_column_name_to_local(self, entity_type: type[Entity]) -> dict[str, str]:
        relation = self._get_relation(entity_type)
        return {remote_column.name: local_column.name for local_column, remote_column in relation.local_remote_pairs}

    def _get_relation(self, entity_type: type[Entity]) -> sqlalchemy.orm.Relationship:
        if (relation := entity_type.get_key_to_relation().get(self.key)) is None:
            raise KeyError(f"Not found relation {self.key!r} in {entity_type.__name__}"
                           f"({entity_type.get_relation_names()})")
        return relation

    def subquery(
            self,
            attrs: list[Expression] | None = None,
            vars: dict[str, Expression] | None = None,
            filters: list[Expression] | None = None,
            searches: list[Expression] | None = None,
            orders: list[Order] | None = None,
            limit: int | None = None,
            offset: int | None = None,
    ):
        from .subquery import SubQuery
        return SubQuery(
            over=self, attrs=attrs or [], vars=vars or {},
            filters=filters or [], searches=searches or [],
            orders=orders or [], limit=limit, offset=offset
        )

    def exists(
            self,
            vars: dict[str, Expression] | None = None,
            filters: list[Expression] | None = None,
            searches: list[Expression] | None = None,
    ):
        from .function import Function
        subquery = self.subquery(vars=vars, filters=filters, searches=searches)
        return Function(key="exists", args=[subquery])

    def shortcut(self) -> str:
        return f"rel.{self.key}"

==> entity_query_abm-2.0.70780a0.tar.gz: entity_query_abm-2.0.70780a0/entity_query/relation.py <==
#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import functools
from dataclasses import dataclass, field
from typing import TypeVar, Self, Generic

import sqlalchemy.orm
from entity_read.entity import Entity

from .expression import Expression
from .order import Order
from .query_token import QueryToken
from .remote_entity import RemoteEntity


EntityType = TypeVar('EntityType', bound=RemoteEntity, covariant=True)


@dataclass(frozen=True, kw_only=True, repr=False)
class Relation(QueryToken, Generic[EntityType]):
    key: str
    parent: RemoteEntity | None = field(hash=False, default=None)

    def __call__(self, parent: RemoteEntity | None) -> Self:
        target_type: type[Relation] = Relation
        if self._type_arg:
            target_type = Relation[self._type_arg] # type: ignore[index]
        result: Relation = target_type(key=self.key, parent=parent)
        return result

    def __str__(self) -> str:
        cls = type(self)
        return (f'{cls.__name__}{f"[{self._entity.__name__}]" if self._entity else ""}'
                f'(parent={self.parent},key={self.key!r})')

    @functools.cached_property
    def _type_arg(self) -> type[EntityType]:
        generic_cls = getattr(self, '__orig_class__', None)
        type_args = getattr(generic_cls, '__args__', None)
        return type_args[0] if type_args else None

    @functools.cached_property
    def _entity(self) -> type[EntityType]:
        return self._type_arg

    @functools.cached_property
    def related(self) -> EntityType:
        return self._type_arg(parent=self)

    __repr__ = __str__

    def get_related_entity(self, entity_type: type[Entity]) -> type[Entity]:
        relation = self._get_relation(entity_type)
        result = relation.entity.entity
        return result

    def get_local_column_name_to_remote(self, entity_type: type[Entity]) -> dict[str, str]:
        relation = self._get_relation(entity_type)
        return {local_column.name: remote_column.name for local_column, remote_column in relation.local_remote_pairs}

    def get_remote_column_name_to_local(self, entity_type: type[Entity]) -> dict[str, str]:
        relation = self._get_relation(entity_type)
        return {remote_column.name: local_column.name for local_column, remote_column in relation.local_remote_pairs}

    def _get_relation(self, entity_type: type[Entity]) -> sqlalchemy.orm.Relationship:
        if (relation := entity_type.get_key_to_relation().get(self.key)) is None:
            raise KeyError(f"Not found relation {self.key!r} in {entity_type.__name__}"
                           f"({entity_type.get_relation_names()})")
        return relation

    def subquery(
            self,
            attrs: list[Expression] | None = None,
            vars: dict[str, Expression] | None = None,
            filters: list[Expression] | None = None,
            searches: list[Expression] | None = None,
            orders: list[Order] | None = None,
            limit: int | None = None,
            offset: int | None = None,
    ):
        from .subquery import SubQuery
        return SubQuery(
            over=self, attrs=attrs or [], vars=vars or {},
            filters=filters or [], searches=searches or [],
            orders=orders or [], limit=limit, offset=offset
        )

    def exists(
            self,
            vars: dict[str, Expression] | None = None,
            filters: list[Expression] | None = None,
            searches: list[Expression] | None = None,
    ):
        from .function import Function
        subquery = self.subquery(vars=vars, filters=filters, searches=searches)
        return Function(key="exists", args=[subquery])

    def shortcut(self) -> str:
        return f"rel.{self.key}"

    def shortcut_over(self) -> str:
        return self.shortcut()

    def get_parent(self) -> RemoteEntity | None:
        return self.parent
